import matplotlib.pyplot as plt
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


def main(do_plot=False):
    t = Timer(op_horizon=24, step_size=3600, initial_date=(2018, 3, 15), initial_time=(0, 0, 0))
    w = Weather(timer=t, location=(50.76, 6.07))
    p = Prices(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    fi = FixedLoad(environment=e, method=1, annual_demand=3000.0, profile_type="H0")
    pv = Photovoltaic(environment=e, method=1, peak_power=6.0)
    ba = Battery(environment=e, e_el_max=8.4, p_el_max_charge=3.6, p_el_max_discharge=3.6)

    plot_time = list(range(t. timesteps_used_horizon))
    fig, axs = plt.subplots(1, 3)
    axs[0].plot(plot_time, p. da_prices, color="black")
    axs[0].set_title("Day-ahead energy market prices [ct/kWh]")
    axs[1].plot(plot_time, fi.p_el_schedule, color="black")
    axs[1].set_title("Single-family house electrical load demand [kW]")
    axs[2].plot(plot_time, pv.p_el_supply, color="black")
    axs[2].set_title("Residential photovoltaics generation [kW]")
    for ax in axs.flat:
        ax.set(xlabel="Time [h]", xlim=[0, t.timesteps_used_horizon-1])
    plt.grid()

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return

    bd = Building(environment=e, objective="none")
    bes = BuildingEnergySystem(environment=e)
    ap = Apartment(environment=e)
    bd. addMultipleEntities(entities=[bes, ap])
    bes . addDevice(objectInstance=pv)
    ap. addMultipleEntities(entities=[fi, ba])

    cd = CityDistrict(environment=e, objective="price")
    cd. addEntity(bd, position=(0, 0))

    opt = CentralOptimization(city_district=cd, mode="integer")
    res = opt.solve()
    cd.copy_schedule(dst="optim_schedule")

    from pycity_scheduling.util.metric import self_consumption
    from pycity_scheduling.util.plot_schedules import plot_entity
    from pycity_scheduling.util.write_schedules import schedule_to_json

    cd.load_schedule(schedule="optim_schedule")

    plot_entity(entity=cd, schedule=["optim_schedule"], title="City district - Cost-optimal schedules")
    plot_entity(entity=ba, schedule=["optim_schedule"], title="Battery unit - Cost-optimal schedules")

    print(self_consumption(entity=bd))

    schedule_to_json(input_list=[fi, pv, ba], file_name="cost_optim.json", schedule=["optim_schedule"])


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a very simple power scheduling example using the stand-alone (reference) optimization algorithm.


def main(do_plot=False):
    print("\n\n------ Example 03: Algorithm Stand-Alone ------\n\n")

    # Define timer, price, weather, and environment objects:
    t = Timer(op_horizon=2, step_size=3600)
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "peak-shaving":
    cd = CityDistrict(environment=e, objective='peak-shaving')

    # Schedule two sample buildings. The buildings' objectives are defined as "price".

    # Building no. one comes with fixed load, space heating, electrical heater, pv unit, thermal energy storage, and
    # electrical energy storage:
    bd1 = Building(environment=e, objective='price')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(environment=e)
    bd1.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
    bes.addDevice(ths)
    eh = ElectricalHeater(environment=e, p_th_nom=10)
    bes.addDevice(eh)
    ap = Apartment(environment=e)
    bd1.addEntity(ap)
    load = np.array([10.0, 10.0])
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    pv = Photovoltaic(environment=e, method=1, peak_power=4.6)
    bes.addDevice(pv)
    bat = Battery(environment=e, e_el_max=4.8, p_el_max_charge=3.6, p_el_max_discharge=3.6)
    bes.addDevice(bat)

    # Building no. two comes with deferrable load, curtailable load, space heating, chp unit, thermal energy storage and
    # an electrical vehicle:
    bd2 = Building(environment=e, objective='price')
    cd.addEntity(entity=bd2, position=[0, 0])
    bes = BuildingEnergySystem(environment=e)
    bd2.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=35, soc_init=0.5)
    bes.addDevice(ths)
    chp = CombinedHeatPower(environment=e, p_th_nom=20.0)
    bes.addDevice(chp)
    ap = Apartment(environment=e)
    bd2.addEntity(ap)
    load = np.array([20.0, 20.0])
    dl = DeferrableLoad(environment=e, p_el_nom=2.0, e_consumption=2.0, load_time=[1, 1])
    ap.addEntity(dl)
    cl = CurtailableLoad(environment=e, p_el_nom=1.6, max_curtailment=0.8)
    ap.addEntity(cl)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    ev = ElectricalVehicle(environment=e, e_el_max=37.0, p_el_max_charge=22.0, soc_init=0.65, charging_time=[0, 1])
    ap.addEntity(ev)

    # Perform the scheduling:
    opt = StandAlone(city_district=cd)
    results = opt.solve()
    cd.copy_schedule("stand-alone")

    # Print the building's schedules:
    print("Schedule building no. one:")
    print(list(bd1.p_el_schedule))
    print("Schedule building no. two:")
    print(list(bd2.p_el_schedule))
    print("Schedule of the city district:")
    print(list(cd.p_el_schedule))
    return


# Conclusions:
# If the stand-alone optimization algorithm is applied, the two buildings are scheduled in a way so that they only
# satisfy their local constraints . Local flexibility is neither used to achieve local or system level objectives.
# For this reason, it should be used for reference scheduling purposes only to analyze the 'state-of-the-art' case
# without any energy system optimization performed.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes import (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
from pycity_scheduling.util import extract_pyomo_values
from pycity_scheduling.algorithms.algorithm import IterationAlgorithm, DistributedAlgorithm, SolverNode
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class DualDecomposition(IterationAlgorithm, DistributedAlgorithm):
    """Implementation of the Dual Decomposition Algorithm.

    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex",
                 eps_primal=0.1, rho=2.0, max_iterations=10000, robustness=None):
        """
        Set up the Dual Decomposition Algorithm for optimizing a specific city district.

        Parameters
        ----------
        city_district : CityDistrict
        solver : str, optional
            Solver to use for solving (sub)problems.
        solver_options : dict, optional
            Options to pass to calls to the solver. Keys are the name of
            the functions being called and are one of `__call__`, `set_instance_`,
            `solve`.
            `__call__` is the function being called when generating an instance
            with the pyomo SolverFactory.  Additionally to the options provided,
            `node_ids` is passed to this call containing the IDs of the entities
            being optimized.
            `set_instance` is called when a pyomo Model is set as an instance of
            a persistent solver. `solve` is called to perform an optimization. If
            not set, `save_results` and `load_solutions` may be set to false to
            provide a speedup.
        mode : str, optional
            Specifies which set of constraints to use.
            - `convex`  : Use linear constraints
            - `integer`  : May use non-linear constraints
        eps_primal : float, optional
            Primal stopping criterion for the dual decomposition algorithm.
        rho : float, optional
            Stepsize for the dual decomposition algorithm.
        max_iterations : int, optional
            Maximum number of ADMM iterations.
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        """
        super(DualDecomposition, self).__init__(city_district, solver, solver_options, mode)
        self.eps_primal = eps_primal
        self.rho = rho
        self.max_iterations = max_iterations
        # create solver nodes for each entity
        self.nodes = [
            SolverNode(solver, solver_options, [entity], mode, robustness=robustness)
            for entity in self.entities
        ]
        # create pyomo parameters for each entity
        for node, entity in zip(self.nodes, self.entities):
            node.model.beta = pyomo.Param(mutable=True, initialize=1)
            node.model.lambdas = pyomo.Param(entity.model.t, mutable=True, initialize=0)
        self._add_objective()

    def _add_objective(self):
        for i, node, entity in zip(range(len(self.entities)), self.nodes, self.entities):
            obj = node.model.beta * entity.get_objective()
            if i == 0:
                # penalty term is expanded and constant is omitted
                # invert sign of p_el_schedule and p_el_vars (omitted for quadratic
                # term)
                obj -= pyomo.sum_product(node.model.lambdas, entity.model.p_el_vars)
            else:
                obj += pyomo.sum_product(node.model.lambdas, entity.model.p_el_vars)
            node.model.o = pyomo.Objective(expr=obj)
        return

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)

        for node, entity in zip(self.nodes, self.entities):
            node.model.beta = self._get_beta(params, entity)
            if full_update:
                node.full_update(robustness)
        results["r_norms"] = []
        results["lambdas"] = []
        return results, params

    def _is_last_iteration(self, results, params):
        return results["r_norms"][-1] <= self.eps_primal

    def _iteration(self, results, params, debug):
        super(DualDecomposition, self)._iteration(results, params, debug)
        op_horizon = self.entities[0].op_horizon
        if "lambdas" not in params:
            params["lambdas"] = np.zeros(op_horizon)
        lambdas = params["lambdas"]

        # -----------------
        # 1) optimize all entities
        # -----------------
        for i, node, entity in zip(range(len(self.nodes)), self.nodes, self.entities):
            if not isinstance(
                    entity,
                    (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
            ):
                continue
            for t in range(op_horizon):
                node.model.lambdas[t] = lambdas[t]
            node.obj_update()
            node.solve(variables=[entity.model.p_el_vars[t] for t in range(op_horizon)], debug=debug)

        # --------------------------
        # 2) incentive signal update
        # --------------------------
        p_el_schedules = np.array([extract_pyomo_values(entity.model.p_el_vars, float) for entity in self.entities])
        lambdas -= self.rho * p_el_schedules[0]
        lambdas += self.rho * np.sum(p_el_schedules[1:], axis=0)

        # ------------------------------------------
        # 3) Calculate parameters for stopping criteria
        # ------------------------------------------
        r = np.zeros(op_horizon)
        r -= p_el_schedules[0]
        r += np.sum(p_el_schedules[1:], axis=0)

        # save parameters for another iteration
        results["r_norms"].append(np.linalg.norm(r, np.inf))
        results["lambdas"].append(np.copy(lambdas))
        return



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.electrical_heater as eh

from pycity_scheduling.util.generic_constraints import LowerActivationLimit
from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating
from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class ElectricalHeater(ThermalEntityHeating, ElectricalEntity, eh.ElectricalHeater):
    """
    Extension of pyCity_base class ElectricalHeater for scheduling purposes.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
        Common to all other objects. Includes time and weather instances.
    p_th_nom : float
        Nominal thermal power output in [kW].
    eta : float, optional
        Efficiency of the electrical heater. Defaults to one.
    lower_activation_limit : float, optional (only adhered to in integer mode)
        Must be in [0, 1]. Lower activation limit of the electrical heater
        as a percentage of the rated power. When the electrical heater is
        in operation, its power must be zero or between the lower activation
        limit and its rated power.

        - `lower_activation_limit = 0`: Linear behavior
        - `lower_activation_limit = 1`: Two-point controlled

    Notes
    -----
    - EHs offer sets of constraints for operation. In the `convex` mode the
      following constraints and bounds are generated by the EH:

    .. math::
        0 \\geq p_{th\\_heat} &\\geq& -p_{th\\_nom} \\\\
        \\eta * p_{el} &=& - p_{th\\_heat}

    - See also:
        - pycity_scheduling.util.generic_constraints.LowerActivationLimit: Generates additional constraints for the
          `lower_activation_limit` in `integer` mode.
    """

    def __init__(self, environment, p_th_nom, eta=1, lower_activation_limit=0):
        # Flow temperature of 55 C
        super().__init__(environment, p_th_nom*1000, eta, 85, lower_activation_limit)
        self._long_id = "EH_" + self._id_string
        self.p_th_nom = p_th_nom
        self.activation_constr = LowerActivationLimit(self, "p_th_heat", lower_activation_limit, -p_th_nom)

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel.

        Call parent's `populate_model` method and set thermal variables upper
        bounds to `self.p_th_nom`. Also add constraint to bind electrical
        demand to thermal output.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or "integer":
            m.p_th_heat_vars.setlb(-self.p_th_nom)
            m.p_th_heat_vars.setub(0.0)

            def p_coupl_rule(model, t):
                return - model.p_th_heat_vars[t] == self.eta * model.p_el_vars[t]
            m.p_coupl_constr = pyomo.Constraint(m.t, rule=p_coupl_rule)

            self.activation_constr.apply(m, mode)

        else:
            raise ValueError(
                "Mode %s is not implemented by class ElectricalHeater." % str(mode)
            )
        return



import matplotlib.pyplot as plt
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


def main(do_plot=False):
    t = Timer(op_horizon=24, step_size=3600, initial_date=(2018, 3, 15), initial_time=(0, 0, 0))
    w = Weather(timer=t, location=(50.76, 6.07))
    p = Prices(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    fi = FixedLoad(environment=e, method=1, annual_demand=3000.0, profile_type="H0")
    pv = Photovoltaic(environment=e, method=1, peak_power=6.0)
    ba = Battery(environment=e, e_el_max=8.4, p_el_max_charge=3.6, p_el_max_discharge=3.6)

    plot_time = list(range(t. timesteps_used_horizon))
    fig, axs = plt.subplots(1, 3)
    axs[0].plot(plot_time, p. da_prices, color="black")
    axs[0].set_title("Day-ahead energy market prices [ct/kWh]")
    axs[1].plot(plot_time, fi.p_el_schedule, color="black")
    axs[1].set_title("Single-family house electrical load demand [kW]")
    axs[2].plot(plot_time, pv.p_el_supply, color="black")
    axs[2].set_title("Residential photovoltaics generation [kW]")
    for ax in axs.flat:
        ax.set(xlabel="Time [h]", xlim=[0, t.timesteps_used_horizon-1])
    plt.grid()

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.city_district as cd

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class CityDistrict(ElectricalEntity, cd.CityDistrict):
    """
    Extension of pyCity_base class CityDistrict for scheduling purposes. Also represents the district operator.

    Parameters
    ----------
    environment : Environment
    objective : str, optional
        Objective for the district operator. Default is 'price'.

        - 'price' : Optimize for the minimum total cost given by `prices.da_prices`.
        - 'peak-shaving' : Try to 'flatten' the schedule as much as possible.
        - 'co2' : Optimize for the minimum total co2 emissions given by `prices.co2_prices`.
        - 'valley-filling' : Try to fill the 'valleys' given by a reference power profile.
        - 'max-consumption' : Try to minimize the maximum power subscription.
        - 'self-consumption' : Try to maximize the self-consumption of the local power generation.
        - 'flexibility-quantification' : To be used to quantify the flexibility potential of the city district only.
        - 'none' : No objective.
    valley_profile : numpy.ndarray, optional
        Profile to be filled by applying the valley filling objective.

    Notes
    -----
    - The constraints generated by a CD are the same as the ones created by an EntityContainer.
    """

    def __init__(self, environment, objective='price', valley_profile=None):
        super().__init__(environment)
        self._long_id = "CD_" + self._id_string
        self.objective = objective
        self.valley_profile = valley_profile

    def populate_model(self, model, mode="convex"):
        """
        Add city district block to pyomo ConcreteModel.

        Call parent's `populate_model` methods and set variables lower
        bounds to `None`.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model

        if mode in ["convex", "integer"]:
            m.p_el_vars.setlb(None)
        else:
            raise ValueError(
                "Mode %s is not implemented by city district." % str(mode)
            )
        return

    def get_objective(self, coeff=1):
        if self.objective == 'valley-filling':
            e = coeff * pyomo.sum_product(self.model.p_el_vars, self.model.p_el_vars)
            valley = self.valley_profile[self.op_slice]
            e += 2 * coeff * pyomo.sum_product(valley, self.model.p_el_vars)
            return e
        elif self.objective == 'price':
            prices = self.environment.prices.da_prices[self.op_slice]
            s = sum(abs(prices))
            if s > 0:
                prices = prices * self.op_horizon / s
                return pyomo.sum_product(prices, self.model.p_el_vars)
            else:
                return 0
        else:
            return super().get_objective(coeff)

    def get_lower_entities(self):
        for node in self.nodes.values():
            yield node['entity']

    def account_imbalance(self):
        """
        Changes the current schedule to account imbalances.

        The imbalances are determined from the current schedules of the contained
        entities.
        """

        total = np.zeros_like(self.schedule["p_el"])
        for e in self.get_lower_entities():
            total += e.schedule["p_el"]

        self.schedule["p_el"] = total
        return



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
from warnings import warn

from pycity_scheduling.classes.battery import Battery
from pycity_scheduling import util


class ElectricalVehicle(Battery):
    """
    Class representing an electrical vehicle for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    e_el_max : float
        Electric capacity of the battery in [kWh].
    p_el_max_charge : float
        Maximum charging power in [kW].
    p_el_max_discharge : float, optional
        Maximum discharging power in [kW]. Defaults to zero.
    soc_init : float, optional
        Initial state of charge. Defaults to 50%.
    charging_time : array of binaries, optional
        Indicator when electrical vehicle can be charged.

        - `charging_time[t] == 0`: EV cannot be charged in t
        - `charging_time[t] == 1`: EV can be charged in t

        It must contain at least one `0` otherwise the model will become
        infeasible. Its length has to be consistent with `ct_pattern`.
        Defaults to only charge during night.
    ct_pattern : str, optional
        Define how the `charging_time` profile is to be used.

        - `None` : Profile matches simulation horizon (default).
        - 'daily' : Profile matches one day.
        - 'weekly' : Profile matches one week.

    Notes
    -----
    - EVs offer sets of constraints for operation. The :math:`e_{el}` equivalence
      constraint is replaced by the following constraint:

    .. math::
        e_{el} &=& e_{el\\_previous} + (\\eta * p_{el\\_demand}
        - (1 / \\eta) * p_{el\\_supply} - p_{el\\_drive}) * \\Delta t \\\\

    - The following constraints are added:

    .. math::
        p_{el\\_drive} \\geq 0 \\\\
        p_{el\\_demand\\_i} = p_{el\\_supply} = 0,
        & \\quad \\text{if} \\quad ct\\_pattern_i = 0 \\\\
        p_{el\\_drive\\_i} = 0, & \\quad \\text{if} \\quad ct\\_pattern_i = 1 \\\\
        e_{el\\_i} = soc\\_init * e_{el\\_max}, & \\quad \\text{if} \\quad \\sum_{j=0}^i ct\\_pattern_j = 0 \\\\
        e_{el\\_i} = 0.2 * e_{el\\_max}, & \\quad \\text{else if} \\quad ct\\_pattern_i = 0 \\\\
        e_{el\\_i} = e_{el\\_max}, & \\quad \\text{else if} \\quad ct\\_pattern_i+1 = 0

    - The constraint for the parameter `storage_end_equality` is removed. Instead,
      the EV needs to be fully charged at the end of the `simu_horizon` if parameter `ct_pattern`
      is one at the end of the simulation horizon.
    """

    def __init__(self, environment, e_el_max, p_el_max_charge, p_el_max_discharge=0.0,
                 soc_init=0.5, charging_time=None, ct_pattern=None):
        super().__init__(environment, e_el_max, p_el_max_charge, p_el_max_discharge, soc_init, eta=1,
                         storage_end_equality=False)
        self._kind = "electricalvehicle"
        self._long_id = "EV_" + self._id_string

        self.charging_time_initial = charging_time
        self.ct_pattern = ct_pattern

        if charging_time is None:
            # load at night, drive during day
            ts_per_day = int(24 / self.time_slot)
            a = int(ts_per_day / 4)
            b = int(ts_per_day / 2)
            c = ts_per_day - a - b
            charging_time = [1] * a + [0] * b + [1] * c
            self.ct_pattern = 'daily'
        self.charging_time = util.compute_profile(self.timer, charging_time, self.ct_pattern)

        # Contains all starting timesteps of charging periods.
        self.starting_timesteps = np.where(np.diff(charging_time, prepend=0, append=0) == 1)[0]
        # Contains all stopping timesteps of charging periods.
        self.stopping_timesteps = np.where(np.diff(charging_time, prepend=0, append=0) == -1)[0]
        # Contains all length of charging periods in timesteps.
        self.charging_durations = self.stopping_timesteps - self.starting_timesteps
        # Links charging periods to indices in previous arrays.
        self.charging_indices = np.zeros_like(charging_time, dtype=np.int)
        for i, start in enumerate(self.starting_timesteps):
            self.charging_indices[start:start + self.charging_durations[i]] = np.full(self.charging_durations[i], i)

        # Contains the amount of energy to charge in charging periods.
        self.required_charges = np.full_like(self.charging_durations, 0.8 * e_el_max, dtype=np.float)

        if len(self.charging_time) > 0:
            # In first charging cycle the required charge is determined by soc_init and is not necessarily 80%.
            self.required_charges[0] = (1 - soc_init) * e_el_max

            if self.charging_time[-1]:
                # Limit charge at end of simu horizon to not become infeasible.
                self.required_charges[-1] = min(self.required_charges[-1], self.charging_durations[-1] *
                                                self.time_slot * p_el_max_charge * self.eta_charge)

        if any(self.required_charges > self.charging_durations * self.time_slot * p_el_max_charge * self.eta_charge):
            warn("Charging pattern results in infeasible constraints.")

        self.new_var("p_el_drive")

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method. Replace coupling
        constraints from Battery class with coupling constraints
        of EV. Simulate power consumption while driving.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or "integer":
            # Simulate power consumption while driving
            m.p_el_drive_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0, None), initialize=0)

            # Replace coupling constraints from Battery class
            m.del_component("e_constr")
            m.del_component("e_end_constr")

            def e_rule(model, t):
                delta = (
                    (self.eta_charge * model.p_el_demand_vars[t]
                     - (1.0 / self.eta_discharge) * model.p_el_supply_vars[t]
                     - model.p_el_drive_vars[t])
                    * self.time_slot
                )
                e_el_last = model.e_el_vars[t - 1] if t >= 1 else model.e_el_init
                return model.e_el_vars[t] == e_el_last + delta
            m.e_constr = pyomo.Constraint(m.t, rule=e_rule)
        else:
            raise ValueError(
                "Mode %s is not implemented by class ElectricalVehicle." % str(mode)
            )
        return

    def update_model(self, mode=""):
        m = self.model

        super().update_model(mode)

        timestep = self.timestep
        charging_time = self.charging_time[self.op_slice]
        # Is true if t is before the initial charging period.
        is_initial = not any(self.charging_time[:timestep])

        # Reset e_el bounds
        for t in self.op_time_vec:
            m.e_el_vars[t].setub(self.e_el_max)
            m.e_el_vars[t].setlb(0.0)

        for t in self.op_time_vec:
            if charging_time[t]:
                m.p_el_demand_vars[t].setub(self.p_el_max_charge)
                m.p_el_supply_vars[t].setub(self.p_el_max_discharge)
                m.p_el_drive_vars[t].setub(0.0)
                is_initial = False
            else:
                m.p_el_demand_vars[t].setub(0.0)
                m.p_el_supply_vars[t].setub(0.0)
                m.p_el_drive_vars[t].setub(None)

                if is_initial:
                    # Remain at soc_init before the first charging period is reached.
                    m.e_el_vars[t].setub(self.soc_init * self.e_el_max)
                    m.e_el_vars[t].setlb(self.soc_init * self.e_el_max)
                else:
                    m.e_el_vars[t].setub(0.2 * self.e_el_max)
                    m.e_el_vars[t].setlb(0.2 * self.e_el_max)

            if t + 1 < self.op_horizon:
                if charging_time[t] and not charging_time[t+1]:
                    # Full battery at end of charging period
                    m.e_el_vars[t].setub(self.e_el_max)
                    m.e_el_vars[t].setlb(self.e_el_max)

        if charging_time[-1]:
            current_ts = (timestep + self.op_horizon - 1) % int(24 / self.time_slot)
            i = self.charging_indices[current_ts]

            first_ts = self.starting_timesteps[i]
            portion_charge = (current_ts - first_ts + 1) * self.required_charges[i] / self.charging_durations[i]
            starting_charge = 0.2 * self.e_el_max if i != 0 else self.soc_init * self.e_el_max
            m.e_el_vars[self.op_horizon-1].setub(starting_charge + portion_charge)
            m.e_el_vars[self.op_horizon-1].setlb(starting_charge + portion_charge)
        return

    def get_objective(self, coeff=1):
        """
        Objective function for entity level scheduling.

        Return the objective function of the electric vehicle weighted with
        coeff. Quadratic term with additional weights to reward charging the
        vehicle earlier.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        m = self.model
        c = np.array(list(map(lambda x: x+1, range(self.op_horizon))))
        c = c * (coeff * self.op_horizon / sum(c))
        return pyomo.sum_product(c, m.p_el_vars, m.p_el_vars)



"""
The pycity_scheduling framework


Copyright (C) 2022,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

import pycity_scheduling.util.factory as factory
import pycity_scheduling.util.debug as debug


# This examples shows the creation of a (complex) city district setup using the pycity_scheduling district generator.
# The obtained district setup is based on building stock information data provided by the EU project TABULA/EPISCOPE
# (https://episcope.eu/welcome/). The functions required for the city district setup can be found in the
# pycity_scheduling factory file.


def main(do_plot=False):
    print("\n\n------ Example 13: District Generator ------\n\n")

    # First, create an environment using the factory's "generate_standard_environment" method. The environment
    # automatically encapsulates time, weather, and price data/information.
    env = factory.generate_standard_environment(initial_date=(2018, 12, 6), step_size=900, op_horizon=96)

    # Create 20 single-family houses:
    num_sfh = 20

    # 50% SFH.2002, 30% SFH.2010, 20% SFH.2016 (based on TABULA):
    sfh_distribution = {
        'SFH.2002': 0.5,
        'SFH.2010': 0.3,
        'SFH.2016': 0.2,
    }

    # 50% of the single-family houses are equipped with heat pump, 10% with boiler, and 40% with electrical heater:
    sfh_heating_distribution = {
        'HP': 0.5,
        'BL': 0.1,
        'EH': 0.4,
    }

    # All single-family houses are equipped with a fixed load, 20% have a deferrable load, and 30% have an electric
    # vehicle. Moreover, 50% of all single-family houses have a battery unit and 80% have a rooftop photovoltaic unit
    # installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    sfh_device_probs = {
        'FL': 1,
        'DL': 0.2,
        'EV': 0.3,
        'BAT': 0.5,
        'PV': 0.8,
    }

    # Create 5 multi-family houses (number of apartments according to TABULA):
    num_mfh = 5

    # 60% MFH.2002, 20% SFH.2010, 20% SFH.2016 (based on TABULA):
    mfh_distribution = {
        'MFH.2002': 0.6,
        'MFH.2010': 0.2,
        'MFH.2016': 0.2,
    }

    # 40% of the multi-family houses are equipped with heat pump, 20% with boiler, and 40% with electrical heater:
    mfh_heating_distribution = {
        'HP': 0.4,
        'BL': 0.2,
        'EH': 0.4,
    }

    # All apartments inside a multi-family houses are equipped with a fixed load, 20% have a deferrable load, and 20%
    # have an electric vehicle. Moreover, 40% of all multi-family houses have a battery unit and 80% have a rooftop
    # photovoltaic unit installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    mfh_device_probs = {
        'FL': 1,
        'DL': 0.2,
        'EV': 0.2,
        'BAT': 0.4,
        'PV': 0.8,
    }

    # Finally, create the desired city district using the factory's "generate_tabula_district" method:
    district = factory.generate_tabula_district(env, num_sfh, num_mfh,
                                                sfh_distribution,
                                                sfh_heating_distribution,
                                                sfh_device_probs,
                                                mfh_distribution,
                                                mfh_heating_distribution,
                                                mfh_device_probs)

    # Hierarchically print the district and all buildings/assets:
    debug.print_district(district, 2)
    return


# Conclusions:
# The pycity_scheduling's factory method "generate_tabula_district" is a simple and fast way to create heterogeneous
# city district setups in an effective way.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)
